#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>
#include <locale.h>

#include "gettext.h"
#define _(String) gettext (String)

#include "sudoku.h"
#include "sudoku_html.h"

#define MYNAME "sudoku"
#define COPYRIGHT _("Copyright (C) 2005-2013 Peter Selinger")

struct key_value_s {
  char *key;
  int value;
  int is_default;
};
typedef struct key_value_s key_value_t;

#define SYM_RANDOM 0x100  /* symbolic value: random symmetry */

static key_value_t symmetries[] = {
  { "none",       SYM_NONE,   0 }, /* no symmetry */
  { "vertical",   SYM_M_V,    0 }, /* reflection, vertical axis */
  { "horizontal", SYM_M_H,    0 }, /* reflection, horizontal axis */
  { "diagonal",   SYM_M_D,    0 }, /* reflection, diagonal */
  { "codiagonal", SYM_M_C,    0 }, /* reflection, codiagonal */
  { "dc",         SYM_M_DC,   0 }, /* reflection, diagonal+codiagonal */
  { "vh",         SYM_M_VH,   0 }, /* reflection, horizontal+vertical axis */
  { "180",        SYM_R_180,  0 }, /* rotation, 180 degrees */
  { "90",         SYM_R_90,   0 }, /* rotation, 90 degrees */
  { "all",        SYM_ALL,    0 }, /* all symmetries */
  { "random",     SYM_RANDOM, 1 }, /* random "pleasing" symmetry */
  { 0, 0 },
};

#define O_TEXT    1
#define O_BIGTEXT 2
#define O_HTML    3

static key_value_t formats[] = {
  { "text",       O_TEXT,     1 },
  { "bigtext",    O_BIGTEXT,  0 },
  { "html",       O_HTML,     0 },
  { 0, 0, 0 },
};

/* look up a value in a key-value list. When no exact match, allow
   first matching prefix. If no match, return nomatch. */
static int lookup(key_value_t *table, char *s, int nomatch) {
  key_value_t *t;

  for (t=table; t->key != NULL; t++) {
    if (!strcasecmp(t->key, s)) {
      return t->value;
    }
  }
  for (t=table; t->key != NULL; t++) {
    if (!strncasecmp(t->key, s, strlen(s))) {
      return t->value;
    }
  }
  return nomatch;
}

/* print list of available option values, starting in column n */
static int optlist(key_value_t *table, FILE *fout, int n) {
  key_value_t *t;
  int l;

  for (t=table; t->key != NULL; t++) {
    l = strlen(t->key)+1;
    if (t->is_default) {
      l += 10;
    }
    if (n+l > 75) {
      fprintf(fout, "\n");
      n=0;
    }
    n += fprintf(fout, "%s%s%s", t->key, t->is_default ? _(" (default)") : "", (t+1)->key ? ", " : ". ");
  }    
  if (n!=0) {
    fprintf(fout, "\n");
  }
  return 0;
}

static int get_default(key_value_t *table, int nomatch) {
  key_value_t *t;
  for (t=table; t->key != NULL; t++) {
    if (t->is_default) {
      return t->value;
    }
  }
  return nomatch;
}

static int parse_pattern(char *p, int (*pattern)[9]) {
  int i, j;

  for (i=0; i<9; i++) {
    for (j=0; j<9; j++) {
      pattern[i][j] = 1;
    }
  }

  i=j=0;
  while (*p) {
    switch (*p) {
    case '0': case '1':
      if (i>=9 || j>=9) {
	return -1;
      }
      pattern[i][j] = (*p=='0') ? 0 : 1;
      j++;
      break;
    case '/': case '\n':
      i++;
      j=0;
      break;
    case ' ': case '\t':
      break;
    default:
      return -1;
      break;
    }
    p++;    
  }
  return 0;
}

static int parse_board(char *p, int (*board)[9]) {
  int i, j;

  for (i=0; i<9; i++) {
    for (j=0; j<9; j++) {
      board[i][j] = 0;
    }
  }

  i=j=0;
  while (*p) {
    switch (*p) {
    case '0': case '1': case '2':
    case '3': case '4': case '5':
    case '6': case '7': case '8':
    case '9':
      if (i>=9 || j>=9) {
	return -1;
      }
      board[i][j] = (*p) - '0';
      j++;
      break;
    case '/': case '\n':
      i++;
      j=0;
      break;
    case ' ': case '\t':
      break;
    default:
      return -1;
      break;
    }
    p++;    
  }
  return 0;
}

static int parse_range(char *p, int *low, int *high) {

  if (*p == '-') {
    *low = 0;
  } else {
    *low = strtoull(p, &p, 0);
    if (*low<0) {
      return -1;
    }
  }
  if (*p != '-') {
    return -1;
  }
  p++;
  if (*p == '\0') {
    *high = -1;
  } else {
    *high = strtoull(p, &p, 0);
    if (*high<0) {
      return -1;
    }
  }
  if (*p != '\0') {
    return -1;
  }
  return 0;
}

/* pick one of the "pleasing" symmetries at random */
static int random_symmetry(void) {
  static const int s[] = {
    SYM_M_DC, SYM_M_VH, SYM_R_180, SYM_ALL,
  };
  return s[rand() % (sizeof(s)/sizeof(*s))];
}

int main1() {
  sudoku_t su;
  int n;
  int i, j;
  int pattern[9][9] = {
    { 0, 0, 1,  0, 0, 0,  0, 1, 0, }, 
    { 0, 0, 1,  0, 1, 0,  0, 1, 0, }, 
    { 0, 1, 1,  0, 0, 1,  0, 0, 0, }, 

    { 0, 0, 0,  1, 0, 1,  0, 0, 0, }, 
    { 0, 1, 1,  0, 1, 0,  1, 1, 0, }, 
    { 0, 0, 0,  1, 0, 1,  0, 0, 0, }, 

    { 0, 0, 0,  1, 0, 0,  1, 1, 0, }, 
    { 0, 1, 0,  0, 1, 0,  1, 0, 0, }, 
    { 0, 1, 0,  0, 0, 0,  1, 0, 0, }, 
  };

  while (1) {
    memset(&su, 0, sizeof(sudoku_t));

    for (i=0; i<9; i++) {
      for (j=0; j<9; j++) {
	if (pattern[i][j]) {
	  su[i][j] = rand() % 9 + 1;
	}
      }
    }
    
    n = sudoku_traverse(su, MODE_UNIQUE, NULL);
    if (n==0) {
      continue;
    }
    sudoku_print(su, stdout);
    fflush(stdout);
    if (n==1) {
      n = sudoku_traverse(su, MODE_BRANCHES, NULL);
      fprintf(stdout, ngettext("1 solution, %d branch\n", "1 solution, %d branches\n", n), n);
      return 0;
    } else {
      fprintf(stdout, ngettext("%d solution\n", "%d solutions\n", n), n);
    }
  }
}

/*

+-------+-------+-------+
|     4 |       |   3   |
|     6 |   8   |   5   |
|   8 9 |     7 |       |
+-------+-------+-------+
|       | 4   8 |       |
|   4 5 |   9   | 1 7   |
|       | 7   6 |       |
+-------+-------+-------+
|       | 9     | 3 2   |
|   1   |   7   | 6     |
|   7   |       | 4     |
+-------+-------+-------+

*/

static int license(FILE *f) {
  fprintf(f, 
_("This program is free software; you can redistribute it and/or modify it\n"
"under the terms of the GNU General Public License as published by the Free\n"
"Software Foundation; either version 2 of the License, or (at your option)\n"
"any later version.\n"
"\n"
"This program is distributed in the hope that it will be useful, but WITHOUT\n"
"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or\n"
"FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for\n"
"more details.\n"
"\n"
"You should have received a copy of the GNU General Public License along\n"
"with this program; if not, write to the Free Software Foundation, Inc.,\n"
"59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.\n")
	  );
  return 0;
}

static int usage(FILE *f) {
  int n;

  fprintf(f, _("Usage: %s [options]\n"), MYNAME);
  fprintf(f, _("Options:\n"));
  fprintf(f, 
	  _(" -h, --help                 - print this help message and exit\n"
	    " -v, --version              - print version info and exit\n"
	    " -l, --license              - print license info and exit\n"
	    " -n, --number <n>           - generate Sudoku number n (default: random)\n"
	    " -f, --format <fmt>         - output format, default is text\n"
	    " -a, --answer               - print answer. -aa to print answer only.\n"
	    " -s, --symmetry <sym>       - set symmetry, default is random\n"
	    " -L, --level [min]-[max]    - request difficult level\n"
	    " -p, --pattern <pattern>    - set pattern\n"
	    " -S, --solve <board>        - solve the given puzzle\n"
	    " -e, --exact                - request to match pattern exactly\n"));
  n = fprintf(f, _("Available symmetries: "));
  optlist(symmetries, f, n);
  n = fprintf(f, _("Available output formats: "));
  optlist(formats, f, n);
  fprintf(f, _("Patterns are specified as in 000111000/001101100/.../000000000.\n"));
  fprintf(f, _("Boards are specified as in 000123000/004506700/.../000000000.\n"));
  return 0;
}

static struct option longopts[] = {
  {"help",          0, 0, 'h'},
  {"version",       0, 0, 'v'},
  {"license",       0, 0, 'l'},
  {"number",        1, 0, 'n'},
  {"format",        1, 0, 'f'},
  {"answer",        0, 0, 'a'},
  {"symmetry",      1, 0, 's'},
  {"level",         1, 0, 'L'},
  {"pattern",       1, 0, 'p'},
  {"solve",         1, 0, 'S'},
  {"exact",         0, 0, 'e'},

  {0, 0, 0, 0}
};

static char *shortopts = "hvln:f:as:L:p:S:e";

struct options_s {
  int sym;                    /* symmetry */
  int have_pattern;           /* have pattern? */
  int pattern[9][9];          /* pattern */
  int solve;                  /* solve? */
  int board[9][9];            /* board to solve */
  int exact;                  /* exactly match pattern? */
  int min_level, max_level;   /* requested difficulty; -1=not set. */
  int format;                 /* O_TEXT, O_HTML, etc */
  int have_seed;              /* random seed requested? */
  unsigned int seed;          /* random seed */
  int answer;                 /* 1=print answer, >1=print only answer */
};
typedef struct options_s options_t;

static int dopts(int ac, char *av[], options_t *opt) {
  int c;
  int s, n;
  char *p;

  /* set defaults */
  opt->sym = get_default(symmetries, SYM_RANDOM);
  opt->have_pattern = 0;
  opt->solve = 0;
  opt->exact = 0;
  opt->min_level = -1;
  opt->max_level = -1;
  opt->format = get_default(formats, O_TEXT);
  opt->have_seed = 0;
  opt->answer = 0;

  while ((c = getopt_long(ac, av, shortopts, longopts, NULL)) != -1) {
    switch (c) {
    case 'h':
      fprintf(stdout, _("%s %s. Creates high-quality sudoku puzzles.\n\n"), MYNAME, VERSION);
      usage(stdout);
      exit(0);
      break;
    case 'v':
      fprintf(stdout, ""MYNAME" "VERSION". %s.\n", COPYRIGHT);
      exit(0);
      break;
    case 'l':
      fprintf(stdout, ""MYNAME" "VERSION". %s.\n\n", COPYRIGHT);
      license(stdout);
      exit(0);
      break;
    case 's':
      s = lookup(symmetries, optarg, -1);
      if (s==-1) {
	fprintf(stderr, _("%s: unknown symmetry: %s.\n"), MYNAME, optarg);
	n = fprintf(stderr, _("Must be one of: "));
	optlist(symmetries, stderr, n);
	exit(1);
      }
      opt->sym = s;
      break;
    case 'n':
      opt->seed = strtoul(optarg, &p, 0);
      if (opt->seed < 0 || *p != '\0') {
        fprintf(stderr, _("%s: invalid number. See --help for more info.\n"), MYNAME);
	exit(1);
      }
      opt->have_seed = 1;
      break;
    case 'a':
      opt->answer++;
      break;
    case 'f':
      s = lookup(formats, optarg, -1);
      if (s==-1) {
	fprintf(stderr, _("%s: unknown format: %s.\n"), MYNAME, optarg);
	n = fprintf(stderr, _("Must be one of: "));
	optlist(formats, stderr, n);
	exit(1);
      }
      opt->format = s;
      break;
    case 'e':
      opt->exact = 1;
      break;
    case 'p':
      s = parse_pattern(optarg, opt->pattern);
      if (s==-1) {
	fprintf(stderr, _("%s: invalid pattern. See --help for more info.\n"), MYNAME);
	exit(1);
      }
      opt->have_pattern = 1;
      break;
    case 'S':
      s = parse_board(optarg, opt->board);
      if (s==-1) {
	fprintf(stderr, _("%s: invalid board. See --help for more info.\n"), MYNAME);
	exit(1);
      }
      opt->solve = 1;
      break;
    case 'L':
      s = parse_range(optarg, &opt->min_level, &opt->max_level);
      if (s==-1) {
        fprintf(stderr, _("%s: invalid level. See --help for more info.\n"), MYNAME);
        exit(1);
      }
      break;
    case '?':
      fprintf(stderr, _("Try --help for more info\n"));
      exit(1);
      break;
    default:
      fprintf(stderr, _("%s: Unimplemented option -- %c\n"), MYNAME, c);
      exit(1);
    }
  }

  return 0;
}


/* process command line options */
int main(int ac, char **av) {
  sudoku_t su, ans;
  int level;
  int sym;
  options_t opt;
  int (*pattern)[9];
  int reduce;
  int r;

  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);

  dopts(ac, av, &opt);

  if (opt.have_seed) {
    srand(opt.seed);
  } else {
    srand(time(0));
  }

  if (opt.solve) {
    r = sudoku_traverse(opt.board, MODE_UNIQUE, NULL);
    if (r == 0) {
      fprintf(stdout, "No solution.\n");
      return 1;
    }
    sudoku_traverse(opt.board, MODE_FILL | FLAG_RANDOM, NULL);
    level = sudoku_traverse(opt.board, MODE_BRANCHES, NULL);
    if (opt.format == O_TEXT) {
      fprintf(stdout, _("Sudoku, Level %d:\n"), level);
      sudoku_print(opt.board, stdout);
    } else if (opt.format == O_BIGTEXT) {
      fprintf(stdout, _("Sudoku, Level %d:\n"), level);
      sudoku_print_big(opt.board, stdout);
    } else {
      print_sudoku_html(opt.board, stdout, level);
    }
    if (r > 1) {
      fprintf(stdout, "Multiple solutions.\n");
    } 
    return 0;
  }

  if (opt.sym == SYM_RANDOM) {
    sym = random_symmetry();
  } else {
    sym = opt.sym;
  }

  if (opt.have_pattern) {
    pattern = opt.pattern;
    reduce = opt.exact ? 0 : 1;
  } else {
    pattern = NULL;
    reduce = 1;
  }

  while (1) {
    level = sudoku(su, ans, pattern, reduce, sym);
    if (level >= opt.min_level && (level <= opt.max_level || opt.max_level == -1)) {
      break;
    }
  }
    
  if (opt.answer < 2) {
    /* output the final solution */
    
    if (opt.format == O_TEXT || opt.format == O_BIGTEXT) {
      if (opt.have_seed) {
	fprintf(stdout, _("Sudoku #%d, Level %d:\n"), opt.seed, level);
      } else {
	fprintf(stdout, _("Sudoku, Level %d:\n"), level);
      }
    }

    if (opt.format == O_TEXT) {
      sudoku_print(su, stdout);
    } else if (opt.format == O_BIGTEXT) {
      sudoku_print_big(su, stdout);
    } else {
      print_sudoku_html(su, stdout, level);
    }
  }
  if (opt.answer > 0) {
    /* output the answer */
  
    if (opt.format == O_TEXT || opt.format == O_BIGTEXT) {
      if (opt.have_seed) {
	fprintf(stdout, _("#%d, Level %d:\n"), opt.seed, level);
      } else {
	fprintf(stdout, _("Sudoku, Level %d:\n"), level);
      }
    }

    if (opt.format == O_TEXT || opt.format == O_BIGTEXT) {
      fprintf(stdout, _("Answer:"));
      sudoku_print_small(ans, stdout);
    } else {
      print_sudoku_html(ans, stdout, level);
    }
  }

  return 0;
}
